@@ -529,8 +529,6 @@ bool WireFormat::ParseAndMergeField(
             return false;
           }
         } else {
-          VerifyUTF8StringNamedField(value.data(), value.length(), PARSE,
-                                     field->full_name());
         }
         if (field->is_repeated()) {
           message_reflection->AddString(message, field, value);
@@ -1002,8 +1000,6 @@ const char* WireFormat::_InternalParseAndMergeField(
             return nullptr;
           }
         } else {
-          VerifyUTF8StringNamedField(value.data(), value.length(), PARSE,
-                                     field->full_name());
         }
       }
       if (field->is_repeated()) {
@@ -1395,8 +1391,6 @@ uint8_t* WireFormat::InternalSerializeField(const FieldDescriptor* field,
                                            WireFormatLite::SERIALIZE,
                                            field->full_name());
         } else {
-          VerifyUTF8StringNamedField(value.data(), value.length(), SERIALIZE,
-                                     field->full_name());
         }
         target = stream->WriteString(field->number(), value, target);
         break;
